#!/usr/bin/env python
#
# Time Drive 0.2.0 - based on duplicity - Encrypted bandwidth efficient 
# backup. Requires Duplicity Version 0.6.04 released August 01, 2009
#
# Copyright 2009 Rob Oakes	<LyX-Devel@oak-tree>

import os
import os.path
import time
import re
import sys

MANUAL = 0
DAY = 1
WEEK = 2
MONTH = 3

import duplicity_interface
import timedrive.backupsettings as backupsettings

from duplicity import dup_time


def find_file(file_name):
	"""Using the image_name, search in the common places. Return the 
	path for the image or None if the image couldn't be found."""
	# Ordered by priority, keep global paths before local paths
	
	current_dir = os.path.abspath(os.path.dirname(__file__))
	common_paths = [os.path.join(sys.prefix, 'local', 'share', 'pixmaps'),
		os.path.join(sys.prefix, "share", "pixmaps"),
		"/usr/local/bin", "/usr/bin",
		os.path.abspath(os.path.join(current_dir, "..", "Resources")),
		os.path.abspath(os.path.join(current_dir, "../.."))]
	for path in common_paths:
		filename = os.path.join(path, file_name)
		if os.access(filename, os.F_OK):
			return filename
	
	return None

def DeleteBackup(stringTime, folder):
	"""Deletes the snapshots as specified by time."""
	
	backupTime = time_String2Int(str(stringTime))
	settings = backupsettings.BackupSettings()
	
	if settings.AmazonS3Backup == True:
		S3_AccessId = settings.AmazonS3_AccessId
		S3_SecretKey = settings.AmazonS3_SecretKey
	else:
		S3_AccessId = None
		S3_SecretKey = None
	
	if settings.AmazonS3_Location == "European":
		S3_BucketType = settings.AmazonS3_Location
	else:
		S3_BucketType = None
	
	archiveUrl, gnuPassphrase = ArchiveUrl_FromSettings(folder, settings)
	operationSuccess = duplicity_interface.delete_backup(archiveUrl, gnuPassphrase, backupTime,
		S3_AccessId, S3_SecretKey, S3_BucketType)
	
	return operationSuccess


def ArchiveUrl_FromSettings(folder, settings):
	"""Creates a valid archive Url from the backupsettings.IncludeList"""
	
	folderName = os.path.basename(os.path.normpath(str(folder.FolderName)))
		
	# Create archiveUrl and gnuPassphrase from General Settings
	if (folder.BackupType == folder.GENERALSETTINGS):
			
		# Local Backup Folder
		if settings.LocalBackup:
			archiveUrl = 'file://' + os.path.normpath(str(settings.LocalBackupDirectory)) + '/' + folderName
			
		# Remote Folder (ssh, ftp, webdav)
		elif settings.RemoteBackup:
			archiveUrl = settings.RemoteProtocol + '://' + settings.RemoteUsername + ':' + settings.RemotePassword + '@' + settings.RemotePath + '/' + folderName
			archiveUrl = str(archiveUrl).replace(' ', '-')
			
		# Amazon S3
		elif settings.AmazonS3Backup:
			archiveUrl = "s3+http://" + os.path.normpath(str(settings.AmazonS3_Bucket)) + '/' + folderName
			archiveUrl = str(archiveUrl).replace(' ', '-')
	
	# Create archiveUrl and GnuPassphras from Advanced Folder Settings
	elif (folder.BackupType == folder.LOCAL)|(folder.BackupType == folder.REMOTE):
				
		if folder.BackupType == folder.LOCAL:
			archiveUrl = 'file://' + folder.LocalBackupDirectory
		elif folder.BackupType == folder.REMOTE:
			archiveUrl = folder.RemoteProtocol + '://' + folder.Username + ':' + folder.Password + '@' + folder.RemotePath
			archiveUrl = str(archiveUrl).replace(' ', '-')
	
	# Retrieve the gnuPassphrase
	if folder.EncryptionKeyEnabled == False:
		gnuPassphrase = settings.Options_EncryptionKey
	else:
		gnuPassphrase = folder.EncryptionKey
	
	return str(archiveUrl), str(gnuPassphrase)


def time_String2Int(time_string):
	"""Convenience funciton for converting raw time strings to integer format."""
	time_int = dup_time.stringtotime(time_string)
	return time_int


def time_GetCurrentTime():
	"""Convenience function which can retrieve the current time"""
	intNow = time.time()
	strNow = time_Int2String(intNow)
	return strNow


def time_Int2String(time_int):
	"""Convenience function for converting time in integer format to raw string."""
	time_string = dup_time.timetostring(time_int)
	return time_string


def time_String2Pretty(time_string):
	"""Convenience funciton for converting raw time strings to a more readable format."""
	time_pretty = dup_time.stringtopretty(time_string)
	return time_pretty


def time_Int2Pretty(time_int):
	"""Convenience function for converting time integers to a more readable format."""
	time_pretty = dup_time.timetopretty(time_int)
	return time_pretty


def verify_ExcludePattern(exclude_pattern):
	"""Verifies the specified exclude_pattern or path to ensure that it doesn't contain
	illegal characters."""
	pattern = str(exclude_pattern[0].toUtf8()).strip()
	
	if len(pattern) == 0:
		return None
	
	if pattern.find(':') >=0:
		raise Exception("Exclude patterns can\'t contain \':\' char !")
		return None
	
	return pattern

def isValidRegExp(regexp):
	try:
    		re.compile(str(regexp))
		return True
	except re.error:
    		print "This RE is invalid"
		return False

def setup_cron(backup_frequency):
	"""Method that creates a listing in the users crontab file so that backups can run 
	unattended.  If using Mac OS X, crontab must be enabled and the user must have a valid
	cron file before running."""
	
	programName = find_file('timedrive-backup')
	os.system("crontab -l | grep -v timedrive-backup | crontab - ")
	
	cron_line = ''
	
	if backup_frequency == MANUAL:	# Manual Backups
		cron_line = ''
	elif backup_frequency == DAY:
		cron_line = 'echo "@daily {cmd}"'
	elif backup_frequency == WEEK:
		cron_line = 'echo "@weekly {cmd}"'
	elif backup_frequency == MONTH:
		cron_line = 'echo "@monthly {cmd}"'
	
	if len(cron_line) > 0:
		cmd = programName + ' > /dev/null 2>&1'
		cmd = 'nice -n 19 ' + cmd
		cron_line = cron_line.replace('{cmd}', cmd)
		os.system("(crontab -l; %s) | crontab -" % cron_line)
